# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from pandas import DataFrame as PdDataFrame
from pyspark.sql.dataframe import DataFrame as SparkDataFrame
from ..dataaccess._open_image_blob_info import OpenImageBlobInfo
from azure.storage.blob import BlockBlobService
from ..environ import RuntimeEnv, SparkEnv, PandasEnv
from multimethods import multimethod
import numpy as np
import os
import pandas as pd
import urllib
import cv2
import requests
from pyspark.sql import functions as F
from pyspark.sql.functions import udf
from pyspark.sql import SparkSession
import json

class ImageDownloader:
    """Open Image Dataset image downloader class."""

    @staticmethod
    def DownloadImages(elements, blobInfo, env, path):
        """
        Download images, used in GetImages function

        :param elements: each element is a tuple, it consists of (urlPrefix, subset/imageId, savePath)
         for example ('https://openimagestorage.blob.core.windows.net/openimagedataset/', 'validation/0001eeaf4aed83f9.jpg', 'tmp/')
        :type elements: iterator
        :param runType: local or spark
        :type runType: str
        """
        containerName = blobInfo.blob_container_name
        blobService = BlockBlobService(
            account_name=blobInfo.blob_account_name,
            sas_token=blobInfo.blob_sas_token.lstrip('?'))
        for relativePath in elements:
            if type(env) == SparkEnv:
                # save image in dbfs, path format : "/dbfs/tmp..." -> tmp folder at dbfs
                imagePath =  os.path.join("/dbfs", path, relativePath.split("/")[-1])
                blobService.get_blob_to_path(containerName, relativePath, imagePath)
            elif type(env) == PandasEnv:
                imagePath = os.path.join(path, relativePath.split("/")[-1])
                if not os.path.exists(imagePath):
                    blobService.get_blob_to_path(containerName, relativePath, imagePath)
            else:
                raise ValueError("Input is invalid")

    @multimethod(OpenImageBlobInfo, PdDataFrame, PdDataFrame, str, dict, str, str)
    def GetImages(self, blobInfo, bbox, urlTable, path, imageDict, storeType, tokens):
        """
        Download images (named by image ids) and a image attribute file (imageid, bb-info-list) into local path 
        or customer azure blob.

        :param urlPrefix: prefix of url
        :type urlPrefix: str
        :param bbox: pandans dataframe contains bounding box information
        :type bbox: PdDataFrame
        :param urlTable: pandas dataframe contains imageID and relative path on azure blob.
        :type urlTable: PdDataFrame
        :param path: download path
        :type path: str
        :param imageDict: dict of imageID and bounding boxID list
        :type imageDict: dict
        :param storeType: local or azureblob.
        :type storeType: str
        :param tokens: for authentication on azure blob.
        :type tokens: str

        """
        if not imageDict:
            raise ValueError("Input can not be empty.")

        imageSavePath = os.path.join(path, "image")
        if not os.path.exists(imageSavePath):
            os.makedirs(imageSavePath)
        imageIds = list(imageDict.keys())

        # annotation file
        bboxLists = [v for imageID in imageIds for v in imageDict.get(imageID)]
        bbox = bbox[bbox.index.isin(bboxLists)].copy()
        bbox.reset_index(inplace = True)
        bbox.loc[:,'BoxIDList'] = bbox.apply(lambda x: {'BoxID':x['BoxID'], 'XMin': x['XMin'],'XMax': x['XMax'],'YMin': x['YMin'],'YMax' : x['YMax'],
                                      'IsOccluded':x['IsOccluded'],'IsTruncated':x['IsTruncated'],'IsGroupOf':x['IsGroupOf'],
                                      'IsInside':x['IsInside'],'Category':x['Category']}, axis=1)
        df = bbox.groupby("ImageID", observed=True)['BoxIDList'].apply(list)
        with open(os.path.join(path, "annotation.json"), "w") as f:
            json.dump(df.reset_index().to_dict(orient='records'),f)

        print('download start...')
        # download images
        elements = []
        for imageId in imageIds:
            relativePath = urlTable.loc[imageId, "Subset"] + "/" + imageId + ".jpg"
            elements.append(relativePath)
        ImageDownloader.DownloadImages(elements, blobInfo, PandasEnv(), imageSavePath)
        print('finish.')

    @multimethod(OpenImageBlobInfo, SparkDataFrame, dict, str, dict, str, str)
    def GetImages(self, blobInfo, bbox, urlTable, path, imageDict, storeType, tokens):
        """
        Download images (named by image ids) and a image attribute file (imageid, categorylist, bb-info-list) into local path 
        or customer azure blob.

        :param urlPrefix: prefix of url
        :type urlPrefix: str
        :param bbox: spark dataframe contains bounding box information
        :type bbox: SparkDataFrame
        :param urlTable: dict contains imageID and relative path on azure blob.
        :type urlTable: dict
        :param path: download path
        :type path: str
        :param imageDict: dict of imageID and bounding boxID list
        :type imageDict: dict
        :param storeType: local or azureblob.
        :type storeType: str
        :param tokens: for authentication on azure blob.
        :type tokens: str

        """
        
        if not imageDict:
            raise ValueError("Input can not be null.")

        # validate input path, starting with / or dbfs: is both valid
        if path.startswith("dbfs:"):
            path = path.split(":")[1][1:]
        elif path.startswith("/"):
            path = path[1:]
        else:
            raise ValueError("The path is invalid")

        spark = SparkSession.builder.getOrCreate()
        # get image list and box list
        imageIds = list(imageDict.keys())
        bboxLists = [v for imageID in imageIds for v in imageDict.get(imageID)]

        # annotaiton file
        bboxListDF = spark.sparkContext.parallelize(bboxLists).map(lambda x : (x, )).toDF(["BoxID"])
        fileterDF = bboxListDF.join(bbox, ["BoxID"], "inner")
        fileterDF = fileterDF.withColumn("BoxList", F.struct(F.col("BoxID"), F.col("XMin"),F.col("XMax"),F.col("YMin"),F.col("YMax"),\
                                F.col("IsOccluded"),F.col("IsTruncated"),F.col("IsGroupOf"),F.col("IsInside"),F.col("Category")))
        fileterDF.groupBy("ImageID").agg(F.collect_list("BoxList").alias("BoxList")).write.json(os.path.join(path, "annoatation"))
                
        # download image
        urlList = []
        for imageId in imageIds:
            url = os.path.join(urlTable.get(imageId), imageId + ".jpg")
            urlList.append(url)
        func = ImageDownloader.DownloadImages
        spark.sparkContext.parallelize(urlList).foreachPartition(lambda x: func(x, blobInfo, SparkEnv(), path))

    @multimethod(PdDataFrame, str, int, list, int, str, str)
    def GetImagesByBatch(self, urlPrefix, imageInfoTable, df, path, partitionSize, imageIds, batchsize, storeType, tokens):
        """
        Return a list of images in image classes for training/testing 
        and at the same time a file described imageInstance list will be created in local or blob
        image class will contain:
        (1) image bits
        (2) image tensors in RGB format and will consider other type of format
        (3) image category list (may use number only, like 0 means car, 1 means human, ...)
        (4) image bb list (each element in the list will be like (xMin, yMin, bbcategory, isTruncated, isOcclued, ...))
        (5) ... 
        :param storeType: local or azureblob.
        :type: str
        :param path: download path.
        :type path: str
        :param tokens: for authentication on azure blob.
        :type tokens: str
        :param partitionSize: ...
        :type partitionSize: int
        :param imageIds: if None, means all images.
        :type imageIds: list
        :param batchsize: download images at the same time when training.
        :type batchsize: int   
        :return: a list of images in image classes for training/testing.
        :rtype: list
        """

        pass


    @multimethod(SparkDataFrame, str, int, list, int, str, str)
    def GetImagesByBatch(self, urlPrefix, imageInfoTable, df, path, partitionSize, imageIds, batchsize, storeType, tokens):
        """
        Return a list of images in image classes for training/testing 
        and at the same time a file described imageInstance list will be created in local or blob
        image class will contain:
        (1) image bits
        (2) image tensors in RGB format and will consider other type of format
        (3) image category list (may use number only, like 0 means car, 1 means human, ...)
        (4) image bb list (each element in the list will be like (xMin, yMin, bbcategory, isTruncated, isOcclued, ...))
        (5) ... 
        :param storeType: local or azureblob.
        :type: str
        :param path: download path.
        :type path: str
        :param tokens: for authentication on azure blob.
        :type tokens: str
        :param partitionSize: ...
        :type partitionSize: int
        :param imageIds: if None, means all images.
        :type imageIds: list
        :param batchsize: download images at the same time when training.
        :type batchsize: int   
        :return: a list of images in image classes for training/testing.
        :rtype: list
        """

        pass
     




# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Define runtime environment classes."""

from multimethods import multimethod
from pandas import DataFrame as PdDataFrame
from pyspark.sql.dataframe import DataFrame as SparkDataFrame


class RuntimeEnv(object):
    """Base class definition of runtime environment."""


class SparkEnv(RuntimeEnv):
    """Class definition of SparkEnv."""


class PandasEnv(RuntimeEnv):
    """Class definition of PandasEnv."""


@multimethod(SparkDataFrame)
def get_environ(data):
    """Get runtime environment of spark."""
    return SparkEnv()


@multimethod(PdDataFrame)
def get_environ(data):
    """Get runtime environment of pandas."""
    return PandasEnv()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Blob info of Open Image Dataset."""


class OpenImageBlobInfo:
    """Blob info of Open Image Datatset."""

    def __init__(self):
        """Initialize Blob Info."""

        self.blob_account_name = "openimagestorage" 
        self.blob_container_name = "openimagedataset"
        self.blob_sas_token = (
            r"?st=2019-07-31T09%3A27%3A58Z&se=2119-08-01T09%3A27%3A00Z&sp=rl&sv=2018-03-28&sr=c"
            # [SuppressMessage("Microsoft.Security", "CS002:SecretInNextLine", Justification="Offline sas token")]
            r"&sig=CQr92AvZjItFDYC4S3ZbE%2Bsj%2F%2BD2W9AmPfKwgb8Dr7E%3D")

        self.table_bbox_path = "/datasetTable/bbox.csv"
        self.table_category_path = "/datasetTable/category.csv"
        self.table_imageInfo_path = "/datasetTable/image-t1.csv"
        self.table_imageBoxNum_path = "/datasetTable/image-t2.csv"
        self.table_imageUrl_path = "/datasetTable/image-t3.csv"
        self.azcopy_win_path = "/AzCopy/azcopy.exe"
        self.azcopy_linux_path = "/AzCopy/azcopy"
        self.url_prefix = "https://openimagestorage.blob.core.windows.net/openimagedataset/"
        


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Enable consuming Azure open datasets into dataframes and enrich customer data."""

from ._open_image import OpenImage

__all__ = [
    'OpenImage',
]



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
import sys

from azureml.telemetry import get_telemetry_log_handler


"""Telemetry utils."""

OPENDATASETS_LOGGER_NAMESPACE = "azureml.opendatasets"
OPENDATASETS_INSTRUMENTATION_KEY = 'f65df220-4460-4269-b954-a11c54f8e611'


def add_appinsights_log_handler(logger):
    handler = get_telemetry_log_handler(
        instrumentation_key=OPENDATASETS_INSTRUMENTATION_KEY,
        component_name=OPENDATASETS_LOGGER_NAMESPACE)
    add_handler(logger, handler)


def add_console_log_handler(logger):
    handler = logging.StreamHandler(sys.stdout)
    add_handler(logger, handler)


def get_opendatasets_logger(name, verbosity=logging.DEBUG):
    logger = logging.getLogger(OPENDATASETS_LOGGER_NAMESPACE).getChild(name)
    logger.propagate = False
    logger.setLevel(verbosity)
    add_appinsights_log_handler(logger)
    add_console_log_handler(logger)
    return logger


def add_handler(logger, handler):
    """
    Add a logger handler and skip if the same type already exists.

    :param logger: Logger
    :param handler: handler instance
    """
    handler_type = type(handler)
    for log_handler in logger.handlers:
        if isinstance(log_handler, handler_type):
            return
    logger.addHandler(handler)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Image common types class."""

import matplotlib.pyplot as plt
from ..environ import SparkEnv, PandasEnv, RuntimeEnv

class ImageAnalyzeResult:
    """the ImageSelector analyze result class."""

    def __init__(self, env: RuntimeEnv, resultDic: dict, target: str=None, edaNameList: list=None):
        """
        Initialize.

        :param env: the runtime environment of the class.
        :type env: RuntimeEnv
        :param resultDic: the analyze result dict.
        :type resultDic: dict
        :param target: the target to show.
        :type target: str
        :param edaNameList: the eda name list.
        :type edaNameList: list
        """
        self.env = env
        self.resultDic = resultDic
        self.target = target
        self.edaNameList = edaNameList

    def show(self, saveName: str=None):
        """
        show the result.

        if the target is 'Luminance' or 'Area', use scatter diagram, else barh diagram.
        :param saveName: the save path of the picture.
        :type saveName: str
        :return: figure if RuntimeEnv is SparkEnv
        :rtype: matplotlib.pyplot.figure
        """

        #close the plot shown before
        plt.close('all')
        fig = plt.figure(dpi=100)
        target = self.target
        if target is None:
            plt.title('There is no data or the edaFilterList is empty.')
            if type(self.env) == PandasEnv:
                plt.show()
            else:
                return fig
        elif target =='Luminance':
            xlabel = 'LuminanceMean'
            ylabel = 'LuminanceStd'
            self.__showAsScatter(target, xlabel, ylabel, saveName)
        elif target=='Area':
            xlabel = 'Width'
            ylabel = 'Height'
            self.__showAsScatter(target, xlabel, ylabel, saveName)
        else:
            self.__showAsBarh(target, self.edaNameList, saveName)
        
    def __showAsBarh(self, target: str, edaNameList: list=None, saveName: str=None, subNumber: int=3, limitClass: int=50):
        """
        show the result as barh diagram.

        :param target: the target to show.
        :type target: str
        :param edaNameList: the eda name list.
        :type edaNameList: list
        :param saveName: the save path of the picture.
        :type saveName: str
        :param subNumber: the subplot number: subNumber*subNumber.
        :type subNumber: int
        :param limitClass: only if data size greater than limitClass, then use subplot.
        :type limitClass: int
        :return: figure if RuntimeEnv is SparkEnv
        :rtype: matplotlib.pyplot.figure
        """
        dic = self.resultDic[target]
        label = list(dic.keys())
        length = len(label)
        y = [i for i in range (length)]
        width = list((dic.values()))
        fig = None
    
        if length > limitClass:
            axCount = subNumber*subNumber
            step = length//axCount
            fig = plt.figure(figsize=(25, 70))
            for i in range(axCount):
                ax = fig.add_subplot(subNumber,subNumber,i+1)
                if i==axCount-1:
                    ax.barh(y=range(length-i*step), width=width[:length-i*step], tick_label=label[:length-i*step], alpha=0.6, facecolor='steelblue', edgecolor='blue')
                else:
                    ax.barh(y=range(step), width=width[length-(i*step+step):length-(i*step)], tick_label=label[length-(i*step+step):length-(i*step)], alpha=0.6, facecolor='steelblue', edgecolor='blue')
            fig.suptitle(','.join(edaNameList), y=0.94)
        else:
            fig = plt.figure(figsize=(10, 5))
            ax = fig.add_subplot(111)
            box = ax.get_position()
            ax.set_position([box.x0*2, box.y0, 0.9-box.x0*2, box.height])
            plt.barh(y, width, tick_label=label, alpha=0.6, facecolor='steelblue', edgecolor='blue')
            plt.title(','.join(edaNameList))
        
        
        if saveName is not None:
            plt.savefig(saveName)

        if type(self.env) == PandasEnv:
            plt.show()
        else:
            return fig
        
    def __showAsScatter(self, target: str, xlabel: str, ylabel: str, saveName: str=None): 
        """
        show the result as scatter diagram.

        :param target: the target to show.
        :type target: str
        :param xlabel: the x label.
        :type xlabel: str
        :param ylabel: the y label.
        :type ylabel: str
        :param saveName: the save path of the picture.
        :type saveName: str
        :return: figure if RuntimeEnv is SparkEnv
        :rtype: matplotlib.pyplot.figure
        """
        dic = self.resultDic[target]
        x = list(dic.keys())
        y = list(dic.values())
        
        fig = plt.figure(dpi=100)
        plt.scatter(x, y, color='steelblue', marker='o', edgecolor='black', alpha=0.5)
        plt.title(target)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        
        if saveName is not None:
            plt.savefig(saveName)
            
        if type(self.env) == PandasEnv:
            plt.show()
        else:
            return fig


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import subprocess
import os
import platform
import numpy as np
from multimethods import multimethod
from pandas import read_csv
from azure.storage.blob import BlockBlobService
from pyspark.sql import SparkSession
from pandas import DataFrame as PdDataFrame
from pyspark.sql.dataframe import DataFrame as SparkDataFrame
from ..environ import SparkEnv, PandasEnv, RuntimeEnv
from ..dataaccess._open_image_blob_info import OpenImageBlobInfo
from pyspark.sql.functions import udf

def GetListFromStr(s:str) -> list:
    """A function to parse str into list.

    :param s: column list str.
    :type s: str.
    :return: the convert result.
    :rtype: list
    """
    return s.strip("[]").replace("'","").split(", ")

def GetCategoryList(str1:str) -> list:
    """A function to parse str into category list.

    :param str1: category list str.
    :type str1: str.
    :return: the convert result.
    :rtype: list
    """
    result = []
    str1 = str1.replace("'","")
    if(len(str1)>2):
        str1 = str1[1:len(str1)-1]
        list1 = str1.split("], ")
        for str2 in list1:
            list2 = str2.strip("[]").split(", ")
            result.append(list2)
    return result

@multimethod(PandasEnv, OpenImageBlobInfo)
def load_annotation_tables(env, blobInfo):
    """
    Load the open image dataset tables.

    :param env: runtime environment.
    :type env: PandasEnv
    :param blobInfo: blob info.
    :type blobInfo: OpenImageBlobInfo
    """

    blob_service = BlockBlobService(
        account_name=blobInfo.blob_account_name,
        sas_token=blobInfo.blob_sas_token.lstrip('?'))

    if platform.system() == 'Windows': 
        azcopy_path = blobInfo.azcopy_win_path
    else:
        azcopy_path = blobInfo.azcopy_linux_path
        
    azcopyFileName = azcopy_path.split('/')[-1]
    
    print('Loading the azcopy executable file...')
    tempPath = os.path.join(os.getcwd(), azcopyFileName)
    blob_service.get_blob_to_path(blobInfo.blob_container_name, azcopy_path.lstrip('/'), tempPath)
    print('finish.')
    
    if platform.system() == 'Linux': 
        prog = subprocess.Popen('chmod +x ./azcopy', shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  
        prog.wait()  
        if prog.returncode: 
            raise Exception('fail to Grant execute authority on the azcopy.')

    print('Loading the tables...')
    table_bbox = load_table_bbox_local(blobInfo.url_prefix, blobInfo.table_bbox_path)
    table_category = load_table_category_local(blobInfo.url_prefix, blobInfo.table_category_path)
    table_imageInfo = load_table_imageInfo_local(blobInfo.url_prefix, blobInfo.table_imageInfo_path)
    table_imageBoxNum = load_table_imageBoxNum_local(blobInfo.url_prefix, blobInfo.table_imageBoxNum_path)
    table_imageUrl = load_table_imageUrl_local(blobInfo.url_prefix, blobInfo.table_imageUrl_path)
    print('Table loading finish.')
    
    #delete the temp file
    os.remove(tempPath)

    return table_bbox, table_category, table_imageInfo, table_imageBoxNum, table_imageUrl

@multimethod(SparkEnv, OpenImageBlobInfo)
def load_annotation_tables(env, blobInfo):
    """
    Load the open image dataset tables.

    :param env: runtime environment.
    :type env: SparkEnv
    :param blobInfo: blob info.
    :type blobInfo: OpenImageBlobInfo
    """
    blob_prefix = 'wasbs://%s@%s.blob.core.windows.net/' % (blobInfo.blob_container_name,blobInfo.blob_account_name)
    spark = SparkSession.builder.getOrCreate()
    spark.conf.set('fs.azure.sas.%s.%s.blob.core.windows.net' % (blobInfo.blob_container_name, blobInfo.blob_account_name),
                   blobInfo.blob_sas_token)
    print('Loading the tables...')
    table_bbox = load_table_bbox_spark(spark, blob_prefix, blobInfo.table_bbox_path)
    table_category = load_table_category_spark(spark, blob_prefix, blobInfo.table_category_path)
    table_imageInfo = load_table_imageInfo_spark(spark, blob_prefix, blobInfo.table_imageInfo_path)
    table_imageBoxNum = load_table_imageBoxNum_spark(spark, blob_prefix, blobInfo.table_imageBoxNum_path)
    table_imageUrl = load_table_imageUrl_spark(spark, blob_prefix, blobInfo.table_imageUrl_path)
    print('Table loading finish.')
    return table_bbox, table_category, table_imageInfo, table_imageBoxNum, table_imageUrl


def load_table_bbox_local(url_prefix: str, table_relative_path: str)-> PdDataFrame:
    """
    Load the bounding box table in local(use azcopy).

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: PdDataFrame
    """

    tableName = table_relative_path.split('/')[-1]
    print('Loading the table {0}...'.format(tableName))
    tempPath = os.path.join(os.getcwd(), tableName)
    url = url_prefix + table_relative_path.lstrip('/')
    if platform.system() == 'Windows':
        cmd = 'azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    else:
        cmd = './azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    #use azcopy to download the table to temp path
    prog = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  
    prog.wait()  
    if prog.returncode: 
        raise Exception('fail to load the table: {0}'.format(tableName))
        
    df = read_csv(tempPath, index_col=0, dtype={'BoxID': object, 'ImageID': 'category', 'XMin': np.uint16, 'YMin': np.uint16, 'XMax': np.uint16, 
                                                'YMax': np.uint16, 'Area': np.uint32, 'IsOccluded': np.int8, 'IsTruncated': np.int8, 
                                                'IsGroupOf': np.int8, 'IsDepiction': np.int8, 'IsInside': np.int8, 
                                                'LuminanceMean': np.uint16, 'LuminanceStd': np.uint16, 'Category': 'category'})
    #delete the temp file
    os.remove(tempPath)
    print('finish.')
    return df

def load_table_bbox_spark(spark: SparkSession, blob_prefix: str, table_relative_path: str)-> SparkDataFrame:
    """
    Load the bounding box table in spark.

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: SparkDataFrame
    """
    return spark.read.csv(blob_prefix + table_relative_path.lstrip('/'), header=True, inferSchema=True).cache()

def load_table_category_local(url_prefix: str, table_relative_path: str)-> PdDataFrame:
    """
    Load the category table in local(use azcopy).

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: PdDataFrame
    """

    tableName = table_relative_path.split('/')[-1]
    print('Loading the table {0}...'.format(tableName))
    tempPath = os.path.join(os.getcwd(), tableName)
    url = url_prefix + table_relative_path.lstrip('/')
    if platform.system() == 'Windows':
        cmd = 'azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    else:
        cmd = './azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    #use azcopy to download the table to temp path
    prog = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  
    prog.wait()  
    if prog.returncode: 
        raise Exception('fail to load the table: {0}'.format(tableName))

    df = read_csv(tempPath, index_col=0, converters={"BoxIDList": GetListFromStr, "ImageIDList": GetListFromStr}, dtype={'Category': object})
    #delete the temp file
    os.remove(tempPath)
    print('finish.')
    return df

def load_table_category_spark(spark: SparkSession, blob_prefix: str, table_relative_path: str)-> SparkDataFrame:
    """
    Load the category table in spark.

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: SparkDataFrame
    """

    category = spark.read.csv(blob_prefix + table_relative_path.lstrip('/'), header=True, inferSchema=True)
    listUdf = udf(lambda x : GetListFromStr(x), 'array<string>')
    category = category.withColumn("BoxIDList", listUdf(category["BoxIDList"]))
    category = category.withColumn("ImageIDList", listUdf(category["ImageIDList"]))
    return category.cache()

def load_table_imageInfo_local(url_prefix: str, table_relative_path: str)-> PdDataFrame:
    """
    Load the image info table in local(use azcopy).

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: PdDataFrame
    """

    tableName = table_relative_path.split('/')[-1]
    print('Loading the table {0}...'.format(tableName))
    tempPath = os.path.join(os.getcwd(), tableName)
    url = url_prefix + table_relative_path.lstrip('/')
    if platform.system() == 'Windows':
        cmd = 'azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    else:
        cmd = './azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    #use azcopy to download the table to temp path
    prog = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  
    prog.wait()  
    if prog.returncode: 
        raise Exception('fail to load the table: {0}'.format(tableName))

    df = read_csv(tempPath, index_col=0, converters={"BoxIDList": GetListFromStr, 'CategoryList': GetCategoryList}, 
                  dtype={'ImageID': object, 'ImageQualityMean': np.uint16, 'ImageQualityStd': np.uint16,
                        'Mode': 'category', 'Width': np.uint16, 'Height': np.uint16,
                        'ImageLuminanceMean': np.uint16, 'ImageLuminanceStd': np.uint16})
    #delete the temp file
    os.remove(tempPath)
    print('finish.')
    return df

def load_table_imageInfo_spark(spark: SparkSession, blob_prefix: str, table_relative_path: str)-> SparkDataFrame:
    """
    Load the image info table in spark.

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: SparkDataFrame
    """
    imageInfoTable = spark.read.csv(blob_prefix + table_relative_path.lstrip('/'), header=True, inferSchema=True)
    listUdf = udf(lambda x : GetListFromStr(x), 'array<string>') 
    nestListUdf = udf(lambda x: GetCategoryList(x), 'array<array<string>>')
    imageInfoTable = imageInfoTable.withColumn("BoxIDList", listUdf(imageInfoTable["BoxIDList"]))
    imageInfoTable = imageInfoTable.withColumn("CategoryList", nestListUdf(imageInfoTable["CategoryList"]))
    return imageInfoTable.cache()

def load_table_imageBoxNum_local(url_prefix: str, table_relative_path: str)-> PdDataFrame:
    """
    Load the imageBoxNum table in local(use azcopy).

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: PdDataFrame
    """

    tableName = table_relative_path.split('/')[-1]
    print('Loading the table {0}...'.format(tableName))
    tempPath = os.path.join(os.getcwd(), tableName)
    url = url_prefix + table_relative_path.lstrip('/')
    if platform.system() == 'Windows':
        cmd = 'azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    else:
        cmd = './azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    #use azcopy to download the table to temp path
    prog = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  
    prog.wait()  
    if prog.returncode: 
        raise Exception('fail to load the table: {0}'.format(tableName))

    df = read_csv(tempPath, index_col=0, converters={"ImageIDList": GetListFromStr}, dtype={'NumofboundingBox': np.uint16})
    #delete the temp file
    os.remove(tempPath)
    print('finish.')
    return df

def load_table_imageBoxNum_spark(spark: SparkSession, blob_prefix: str, table_relative_path: str)-> SparkDataFrame:
    """
    Load the imageBoxNum table in spark.

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: SparkDataFrame
    """
    imageBoxNum = spark.read.csv(blob_prefix + table_relative_path.lstrip('/'), header=True, inferSchema=True)
    listUdf = udf(lambda x : GetListFromStr(x), 'array<string>') 
    imageBoxNum = imageBoxNum.withColumn("ImageIDList", listUdf(imageBoxNum["ImageIDList"]))
    return imageBoxNum

def load_table_imageUrl_local(url_prefix: str, table_relative_path: str)-> PdDataFrame:
    """
    Load the imageUrl table in local(use azcopy).

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: PdDataFrame
    """

    tableName = table_relative_path.split('/')[-1]
    print('Loading the table {0}...'.format(tableName))
    tempPath = os.path.join(os.getcwd(), tableName)
    url = url_prefix + table_relative_path.lstrip('/')
    if platform.system() == 'Windows':
        cmd = 'azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    else:
        cmd = './azcopy copy "{source}" "{dest}"'.format(source=url, dest=tempPath)
    #use azcopy to download the table to temp path
    prog = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  
    prog.wait()  
    if prog.returncode: 
        raise Exception('fail to load the table: {0}'.format(tableName))

    df = read_csv(tempPath, index_col=0, dtype={'ImageID': object, 'Subset': 'category'})
    #delete the temp file
    os.remove(tempPath)
    print('finish.')
    return df

def load_table_imageUrl_spark(spark: SparkSession, blob_prefix: str, table_relative_path: str)-> dict:
    """
    Load the imageUrl table in spark, and convert it to dict

    :param url_prefix: url prefix.
    :type url_prefix: str
    :param table_relative_path: table relative path.
    :type table_relative_path: str
    :return: the table.
    :rtype: dict
    """
    df = spark.read.csv(blob_prefix + table_relative_path.lstrip('/'), header=True, inferSchema=True)
    return df.rdd.map(lambda row : (row[0], row[1])).collectAsMap()



from setuptools import setup
from setuptools import find_packages
 
setup(
    name='testOpendataset',
    version='0.1.7',
    description='SDK for the Open Image Dataset.',
    packages= find_packages(),
    author='zhenwan,t-tozho,t-zhiwan',
    author_email= "zhenwan@microsoft.com,t-tozho@microsoft.com,t-zhiwan@microsoft.com",
    url='https://dev.azure.com/zhenwan/zhenwan_default/_git/AMLProjects',
    install_requires=[
        'azureml-telemetry',
        'azure-storage-blob',
        'multimethods>=1.0.0',
        "numpy>=1.16.0",
        "pandas>=0.24.0",
        'pyspark>=2.4.3',
        'matplotlib >= 3.0.3',
        'opencv-contrib-python >= 4.0.25',
        'spacy>=2.1.8'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ]
)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Open Image Dataset."""

from pandas import DataFrame as PdDataFrame
from pyspark.sql.dataframe import DataFrame as SparkDataFrame
from .dataaccess._open_image_blob_info import OpenImageBlobInfo
from ._utils.image_downloader import ImageDownloader
from ._utils.image_selector import ImageSelector
from ._utils.image_common_types import ImageAnalyzeResult
from .environ import SparkEnv, PandasEnv, RuntimeEnv
from ._utils.image_annotation_table_loader import load_annotation_tables
from ._utils.telemetry_utils import get_opendatasets_logger
from azureml.telemetry.activity import ActivityType, log_activity
import spacy

class OpenImage:
    """Open Image Dataset class."""

    #load the spacy model
    try:
        print('loading the spacy model..')
        import en_core_web_md
        __model_spacy = en_core_web_md.load()
        print('finish.')
    except ImportError:
        print('Downloading language model for the spaCy POS tagger\n'
              "(don't worry, this will only happen once)")
        from spacy.cli import download
        download('en_core_web_md')
        import en_core_web_md
        __model_spacy  = en_core_web_md.load()
        print('finish.')

    __blobInfo = OpenImageBlobInfo()

    def __init__(
            self,
            env: str = 'pandas',
            enable_telemetry: bool = True):
        """
        Initializes an instance of the OpenImage class.

        :param env: the runtime environment of the class.
        :type env: str
        :param enable_telemetry: whether to enable telemetry, disabled for UT only.
        :type enable_telemetry: bool
        """
        env = env.lower()
        if env == 'pandas':
            self.env = PandasEnv()
        elif env == 'spark':
            self.env = SparkEnv()
        else:
            raise ValueError("Input must be 'pandas' or 'spark'")

        self.enable_telemetry = enable_telemetry
        if self.enable_telemetry:
            self.logger = get_opendatasets_logger(__name__)
            self.log_properties = self.get_common_log_properties()
            self.log_properties['env'] = env

        self.__selector = ImageSelector(self.env)
        self.__downLoader = ImageDownloader()

        #load the tables
        if not hasattr(OpenImage, '_OpenImage__table_imageUrl') or \
        (type(self.env) != PandasEnv and type(getattr(OpenImage, '_OpenImage__table_imageUrl')) == PdDataFrame) or \
        (type(self.env) != SparkEnv and type(getattr(OpenImage, '_OpenImage__table_imageUrl')) == SparkDataFrame):
            OpenImage.__table_bbox,  OpenImage.__table_category, OpenImage.__table_imageInfo,\
                OpenImage.__table_imageBoxNum, OpenImage.__table_imageUrl = load_annotation_tables(self.env, self.__blobInfo) 

    def get_common_log_properties(self):
        """Get common log properties."""
        return {}

    def AnalyzeImages(self, imageIDList: list=[], edaFilterList: dict={}) -> ImageAnalyzeResult:
        """
        Analyze images depending on EDA metrics and will return different kind of metrics:
        (1) Grayscale or color image distribution
        (2) Bounding boxes number distribution
        (3) Image size distribution
        (4) Image category distribution
        (5) Image luminance distribution
        (6) Image quality distribution (images with low night noise distribution)
        (7) ...

        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: the analyze result.
        :rtype: ImageAnalyzeResult
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'AnalyzeImages',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.AnalyzeImages(OpenImage._OpenImage__table_imageInfo, OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_category, imageIDList, edaFilterList)
        else:
            return self.__selector.AnalyzeImages(OpenImage._OpenImage__table_imageInfo, OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_category, imageIDList, edaFilterList)

    def AnalyzeImagesByBoundingBoxes(self, boxIDList: list=[], edaFilterList: dict={}) -> ImageAnalyzeResult:
        """
        Analyze bounding boxes depending on EDA metrics and will return different kind of metrics:
        (1) bb-level category distribution
        (2) bb-level attributes distribution like isOccluded bbs distribution, isTruncated bbs distribution, isGrouped bbs distribution, ...
        (3) bb size distribution
        (4) bb luminance distribution
        (5) ...

        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: the analyze result.
        :rtype: ImageAnalyzeResult
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'AnalyzeImagesByBoundingBoxes',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.AnalyzeImagesByBoundingBoxes(OpenImage._OpenImage__table_bbox, boxIDList, edaFilterList)
        else:
            return self.__selector.AnalyzeImagesByBoundingBoxes(OpenImage._OpenImage__table_bbox, boxIDList, edaFilterList)

    def SelectImages(self, imageIDList: list=[], boxIDList: list=[], edaFilterList: dict={}) -> dict:
        """
        Select images
        People can select images on different rules and return image id list:
        1. select images by ids, if none, select all images.
        2. select images by one or multiple combination of the following image level eda metrics: 
            (1) image-level category, 
            (2) image size, 
            (3) bounding box number, 
            (4) bounding box ids, 
            (5) image luminance
            (6) image quality (low night noise)
            (7) Grayscale or color image
            (8) ...

        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: a dict like {ImageID: [bboxID1, bboxID2,...]}.
        :rtype: dict
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'SelectImages',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.SelectImages(OpenImage._OpenImage__table_imageInfo, OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_category, imageIDList, boxIDList, edaFilterList)
        else:
            return self.__selector.SelectImages(OpenImage._OpenImage__table_imageInfo, OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_category, imageIDList, boxIDList, edaFilterList)

    def SelectBboxs(self, boxIDList: list=[], edaFilterList: dict={}) -> list:
        """
        Select bounding boxes (bbs)
        People can select bbs on different rules and bbs id list:
        1. select bbs by ids, if none, select all bbs
        2. select bbs by one or multiple combination of the following bb level eda metrics:
            (1) bb-level category
            (2) bb-level attributes like isOccluded, isTruncated, isGrouped, ...
            (3) bb size
            (4) bb luminace
            (5) ...

        :param box_data: bbox data.
        :type box_data: PdDataFrame
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: a boxID list of the query results.
        :rtype: list
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'SelectBboxs',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.SelectBboxs(OpenImage._OpenImage__table_bbox, boxIDList, edaFilterList)
        else:
            return self.__selector.SelectBboxs(OpenImage._OpenImage__table_bbox, boxIDList, edaFilterList)

    def GetCategories(self) -> list:
        """
        Get all categories.

        :return: a category list.
        :rtype: list
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'GetCategories',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.GetCategories(OpenImage._OpenImage__table_category)
        else:
            return self.__selector.GetCategories(OpenImage._OpenImage__table_category)

    def GetCategories(self) -> list:
        """
        Get all categories.

        :return: a category list.
        :rtype: list
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'GetCategories',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.GetCategories(OpenImage.__table_category)
        else:
            return self.__selector.GetCategories(OpenImage.__table_category)

    def ShowImages(self, path: str, imageIDList: list=[], boxIDList: list=[], limitedNum: int=100):
        """
        Show bounding boxes in images and will download images with drawn bounding boxes to local path. 

        :param path: local save path.
        :type path: str
        :param imageIDList: a list of imageIDs(if None, select image id by boxIDList).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param urlPrefix: the download url prefix.
        :type urlPrefix: str
        :param limitedNum: the download image number.
        :type limitedNum: int
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'ShowImages',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__selector.ShowImages(OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_imageUrl, imageIDList, boxIDList, self.__blobInfo.url_prefix, path, limitedNum)
        else:
            return self.__selector.ShowImages(OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_imageUrl, imageIDList, boxIDList, self.__blobInfo.url_prefix, path, limitedNum)

    def GetImages(self, path: str, imageDict: dict = None, storeType: str = "", tokens: str = ""):
        """
        Download images (named by image ids) and a image attribute file (imageid, categorylist, bb-info-list) into local path 
        or customer azure blob.
        :param storeType: local or azureblob.
        :type storeType: str
        :param path: download path.
        :type path: str
        :param tokens: for authentication on azure blob.
        :type tokens: str
        :param imageIds: image id list you want to download.
        :type imageIds: list
        :param imageLimtationNumberWhenNoIds: to avoid download all images unless customer manually give a None here.
        :type imageLimtationNumberWhenNoIds: int
        :param includeAttributedFile: to download only images or both.
        :type includeAttributedFile: bool
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'GetImages',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__downLoader.GetImages(self.__blobInfo, OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_imageUrl, path, imageDict, storeType, tokens)
        else:
            return self.__downLoader.GetImages(self.__blobInfo, OpenImage._OpenImage__table_bbox, OpenImage._OpenImage__table_imageUrl, path, imageDict, storeType, tokens)
    
    def GetImagesByBatch(self, storeType: str, path: str, tokens: str, partitionSize: int, imageIds: list = None, batchsize: int = 100) -> list:
        """
        Return a list of images in image classes for training/testing 
        and at the same time a file described imageInstance list will be created in local or blob
        image class will contain:
        (1) image bits
        (2) image tensors in RGB format and will consider other type of format
        (3) image category list (may use number only, like 0 means car, 1 means human, ...)
        (4) image bb list (each element in the list will be like (xMin, yMin, bbcategory, isTruncated, isOcclued, ...))
        (5) ... 
        :param storeType: local or azureblob.
        :type: str
        :param path: download path.
        :type path: str
        :param tokens: for authentication on azure blob.
        :type tokens: str
        :param partitionSize: ...
        :type partitionSize: int
        :param imageIds: if None, means all images.
        :type imageIds: list
        :param batchsize: download images at the same time when training.
        :type batchsize: int 
        :return: a list of images in image classes for training/testing.
        :rtype: list
        """
        if self.enable_telemetry:
            with log_activity(
                    self.logger,
                    'GetImagesByBatch',
                    ActivityType.PUBLICAPI,
                    custom_dimensions=self.log_properties):
                return self.__downLoader.GetImagesByBatch(OpenImage.__blobInfo.url_prefix, OpenImage._OpenImage__table_imageInfo, OpenImage._OpenImage__table_imageUrl, storeType, path, tokens, partitionSize, imageIds, batchsize)
        else:
            return self.__downLoader.GetImagesByBatch(OpenImage.__blobInfo.url_prefix, OpenImage._OpenImage__table_imageInfo, OpenImage._OpenImage__table_imageUrl, storeType, path, tokens, partitionSize, imageIds, batchsize)


# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Image selector class."""

from pandas import DataFrame as PdDataFrame
from ..environ import SparkEnv, PandasEnv, RuntimeEnv
from .image_common_types import ImageAnalyzeResult
from collections import OrderedDict
from multimethods import multimethod
import logging
import cv2
import numpy as np
from urllib.request import urlopen
import os
from spacy.lang.en import English

from pyspark.sql.functions import udf,size,explode,concat_ws
from pyspark.sql import SparkSession
from pyspark.sql.dataframe import DataFrame as SparkDataFrame
import pyspark.sql.functions as F

class ImageSelector:
    """different kinds of selection to get the image info."""
    def __init__(self, env: RuntimeEnv):
        """
        Initializes an instance of the ImageSelector class.

        :param env: the runtime environment of the class.
        :type env: RuntimeEnv
        """

        self.env = env

    @staticmethod
    def __Numericfun(x: int, conditList: list):
        """
        convert Integer to String based on condition.

        :param x: Number to convert
        :type x: int 
        :param conditList: condition list
        :type conditList: list 
        """
        result = ''
        for item in conditList:
            if type(item) == int:
                if x == item:
                    result = str(item)
                    break
            else:
                if x >= item[0] and x <= item[1]:
                    result = str(item[0]) + '-' + str(item[1])
                    break
        return result

    @staticmethod
    def showBoxInImage(urlPrefix, path, elements):
        """
        draw bounding boxes on images of each partition, and save them.

        :param urlPrefix: the prefix of url
        :type urlPrefix: string
        :param path: save path
        :type path: string
        :param elements: each element contains: subset/imageId and boxList of the image
        :type elements: iterator
        """
        for element in elements:
            imageSubsetAndId = element[0]
            boxList = element[1]
            imageId = imageSubsetAndId.split("/")[1]
            url = urlPrefix + imageSubsetAndId + ".jpg"
            req = urlopen(url)
            arr = np.asarray(bytearray(req.read()), dtype=np.uint8)
            img = cv2.imdecode(arr, -1)
            for box in boxList:
                xmin = box["XMin"]
                xmax = box["XMax"]
                ymin = box["YMin"]
                ymax = box["YMax"]
                text = box["Category"]
                cv2.rectangle(img, (xmin, ymin), (xmax,ymax), (0,255,0), 2)
                cv2.putText(img, text, (xmin, max(ymin,10)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0,0,255), 1)
            cv2.imwrite(os.path.join("/dbfs" + path, imageId + ".jpg"), img)

    def __GetSynonyms(self, nlp: English, categoryList: list, inputStrList: list, length: int=3) -> list: 
        """
        Get synonyms of the inputStrList.

        :param nlp: spacy model.
        :type nlp: spacy.lang.en.English
        :param categoryList: category list.
        :type categoryList: list
        :param inputStrList: input string list.
        :type inputStrList: list
        :param length: the length of the relevant category list.
        :type length: int
        :return: a category list.
        :rtype: list
        """
        tokens = nlp(' '.join(categoryList))
        result = []

        newCategoryList = [category.lower() for category in categoryList]
        for s in inputStrList:    
            s = s.strip().capitalize()

            if s.lower() in newCategoryList:
                result.append(categoryList[newCategoryList.index(s.lower())])
            else:         
                print('There is no category called "%s", will find relevant categories.' % s)
                if s != "":
                    relList = []
                    txt = nlp(s)
                    if txt and txt.vector_norm:
                        for token in tokens:
                            if(token and token.vector_norm):
                                score = token.similarity(txt)
                                if len(relList) >= length:
                                    index = 0 # in order to contrast 
                                    value = 1.0
                                    for i in range(0,len(relList)):
                                        if relList[i][1] < value:
                                            value = relList[i][1]
                                            index = i
                                    if value < score:
                                        relList[index] = [token.text, score]

                                else:
                                    relList.append([token.text, score])

                        nameList =[li[0] for li in relList]
                        for name in nameList:
                            result.extend([category for category in categoryList if name in category])
        return list(set(result))

    def __validateEDA(self, edaFilterList: dict, numericSelection: list=None, categorySelection: list=None, boolSelection: list=None) -> list:
        """
        The function to check the edaFilterList.

        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :param numericSelection: 'Area', 'Height', 'Width', 'LuminanceStd', 'LuminanceMean', 'BoxNumberPerImage', int or tuple with len=2 in every item in the list.
        :type numericSelection: list
        :param categorySelection: 'Category' and 'Mode', string every item in the list.
        :type categorySelection: list
        :param boolSelection: 'IsTruncated' and etc, bool every item in the list.
        :type boolSelection: list
        :return: a new eda filter list.
        :rtype: list
        """
        if numericSelection:
            numericSelectionLower = [s.lower() for s in numericSelection]   
        if categorySelection:
            categorySelectionLower = [s.lower() for s in categorySelection] 
        if boolSelection:
            boolSelectionLower = [s.lower() for s in boolSelection] 
        
        newEDAList = {}
        if not edaFilterList:
            return None

        if type(edaFilterList) != dict:
            raise ValueError("The edaFilterList must be a dict.")

        for edaName, conditList in edaFilterList.items():
            if type(edaName) != str:
                raise ValueError('The edaName must be a string.')    
            edaName = edaName.strip().lower()
        
            #if eda is luminance and for image, then add image prefix(only bbox has boolSelection)
            if boolSelection is None and edaName in ['luminancestd', 'luminancemean']:
                edaName = 'image'+ edaName
            
            if type(conditList) != list:
                raise ValueError('The filter condition must be a list.')

            if numericSelection and edaName in numericSelectionLower:
                edaName = numericSelection[numericSelectionLower.index(edaName)]
                for condit in conditList:
                    if type(condit) == int or (type(condit) == tuple and len(condit) == 2 and type(condit[0]) == int and type(condit[1]) == int):
                        continue
                    else:
                        raise ValueError("The numeric edaName '%s' condition's element must be int or tuple with len=2." % edaName)
                newEDAList[edaName] = conditList
            
            elif categorySelection and edaName in categorySelectionLower:
                edaName = categorySelection[categorySelectionLower.index(edaName)]
                for condit in conditList:
                    if type(condit) != str:
                        raise ValueError("The category edaName '%s' condition's elements must be string." % edaName)

                if edaName == 'Category':
                    #get category Synonyms
                    from .._open_image import OpenImage
                    categoryList = self.GetCategories(OpenImage._OpenImage__table_category)
                    newConditList = self.__GetSynonyms(OpenImage._OpenImage__model_spacy, categoryList, conditList)
                    newEDAList[edaName] = newConditList
                else:
                    newEDAList[edaName] = conditList
            
            elif boolSelection and edaName in boolSelectionLower:
                edaName = boolSelection[boolSelectionLower.index(edaName)]
                newConditList = []
                for condit in conditList:
                    if type(condit) != bool:
                        raise ValueError("The bool edaName '%s' condition's elements must be bool." % edaName)

                    condit = 1 if condit else 0
                    newConditList.append(condit)
                newEDAList[edaName] = newConditList
            else:
                raise ValueError("The edaName '%s' is invalid." % edaName)
            
        return newEDAList

   
    @multimethod(PdDataFrame, PdDataFrame, PdDataFrame, list, list, dict)
    def SearchImagesOnEDA(self, imageInfoTable, bboxInfoTable, categoryTable, imageIDList, boxIDList, edaFilterList):
        """
        The function to search images by the edaFilterList.

        :param imageInfoTable: image info table.
        :type imageInfoTable: PdDataFrame
        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param categoryTable: category table.
        :type categoryTable: PdDataFrame
        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all boxes).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: search result, eda name list, category list.
        :rtype: PdDataFrame, list, list
        """
        numericSelection = ['ImageLuminanceMean', 'ImageLuminanceStd', 'Width', 'Height', 'Area', 'BoxNumberPerImage']
        categorySelection = ['Mode', 'Category']
    
        #check the edaFilterList
        modifiedEdaFilterList = self.__validateEDA(edaFilterList, numericSelection, categorySelection)

        result = None
        edaNameList = None

        
        if imageIDList and boxIDList :
            result = imageInfoTable[imageInfoTable.index.isin(imageIDList)]
            logging.warning("the imageIDList and boxIDList shouldn't exist together.")
        elif imageIDList and not boxIDList:
            result = imageInfoTable[imageInfoTable.index.isin(imageIDList)]
        
        elif boxIDList and not imageIDList:
            imageIDListBox = bboxInfoTable[bboxInfoTable.index.isin(boxIDList)]['ImageID'].tolist()
            result = imageInfoTable[imageInfoTable.index.isin(imageIDListBox)]
        else:
            #This means user does not give any image ids or bbox ids then search on the imageInfoTable.
            result = imageInfoTable
        
        #add BoxNumber column
        result = result.assign(BoxNumberPerImage = result['BoxIDList'].map(lambda x: len(x)).values)
        #add Area column
        result.eval('Area = Width*Height', inplace=True)
    
        filtermask = ''
        imageIDListCate = []
        categoryList = None
        if modifiedEdaFilterList:
            edaNameList = [edaName for edaName in modifiedEdaFilterList.keys()]
        
            for edaName, condit in modifiedEdaFilterList.items():
                if edaName in numericSelection:
                    if len(condit) > 0:
                        filtermask += '('
                        for item in condit:
                            if type(item) == int:
                                filtermask += '{edaName} == {value} or '.format(edaName=edaName, value=item)
                            else:
                                filtermask += '({edaName} >= {left} and {edaName} <= {right}) or '.format(edaName=edaName, left=item[0], right=item[1])
                        filtermask = filtermask.rstrip('or ')
                        filtermask += ') and '
                elif edaName in categorySelection:
                    #Mode
                    if edaName == categorySelection[0]:
                        filtermask += '{edaName} in {conditList} and '.format(edaName=edaName, conditList=condit)
                    #Category
                    else:    
                        categoryList = condit.copy()
                        imageIDListsCate = categoryTable[categoryTable.index.isin(condit)]['ImageIDList'].tolist()
                        imageIDListCate = list(set().union(*imageIDListsCate))
                        filtermask += 'index in @imageIDListCate and '
                else:
                    raise ValueError('The edaName %s is invalid.' % edaName)
                
            #filteremask sample: Mode in ['RGB'] and (Width == 1024 or (Width >= 0 and Width <= 400)) and 
            #(Height == 1024 or (Height >= 0 and Height <= 400)) and ((BoxNumberPerImage >= 0 and BoxNumberPerImage <= 100)) 
            #and ((ImageLuminanceStd >= 0 and ImageLuminanceStd <= 200))
            filtermask = filtermask.rstrip('and ')
            result = result.query(filtermask)

            if result.shape[0] == 0:
                return result, edaNameList, categoryList
        
            if set(edaNameList)<= {'ImageLuminanceMean', 'ImageLuminanceStd'}:
                result = result[['ImageLuminanceMean', 'ImageLuminanceStd']]
            elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
                result = result[['Width', 'Height']]
            else:
                if categoryList:
                    result = result.assign(Category = result['CategoryList'].map(lambda x: ','.join(set([lis[0] for lis in x]))).values)
                    temp = result.drop(columns=['Category','CategoryList'],axis=1)
                    result = temp.join(result['Category'].str.split(',',expand=True).stack().reset_index(level=1,drop=True).rename('Category'))
                    result = result[result['Category'].isin(categoryList)]
                result = result[edaNameList]
                for edaName, condit in modifiedEdaFilterList.items():
                    if edaName in numericSelection:
                        result.loc[:, edaName] = result[edaName].map(lambda x: self.__Numericfun(x, condit))
    
        return result, edaNameList, categoryList
    
    @multimethod(SparkDataFrame, SparkDataFrame, SparkDataFrame, list, list, dict)
    def SearchImagesOnEDA(self, imageInfoTable, bboxInfoTable, categoryTable, imageIDList, boxIDList, edaFilterDict):
        """
        The function to search images by the edaFilterList.

        :param imageInfoTable: image info table.
        :type imageInfoTable: SparkDataFrame
        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param categoryTable: category table.
        :type categoryTable: SparkDataFrame
        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all boxes).
        :type boxIDList: list
        :param edaFilterDict: a dict of eda filters[(edaname, filtercondition)].
        :type edaFilterDict: dict
        :return: search result, eda name list, category list.
        :rtype: SparkDataFrame, list, list
        """
        numericSelection = ['ImageLuminanceMean', 'ImageLuminanceStd', 'Width', 'Height', 'Area', 'BoxNumberPerImage']
        categorySelection = ['Mode', 'Category']
        spark = SparkSession.builder.getOrCreate()
        #check the edaFilterList
    
        modifiedEdaFilterDict = self.__validateEDA(edaFilterDict, numericSelection, categorySelection)
        result = None
        edaNameList = None
    
        if imageIDList and boxIDList:
            temp = spark.sparkContext.parallelize(imageIDList).map(lambda x : (x, )).toDF(["ImageID"])
            result = temp.join(imageInfoTable, ["ImageID"], "inner")
            logging.warning("The imageIDList and boxIDList shouldn't exist together.")
        elif imageIDList and not boxIDList:
            temp = spark.sparkContext.parallelize(imageIDList).map(lambda x : (x, )).toDF(["ImageID"])
            result = temp.join(imageInfoTable, ["ImageID"], "inner")
        elif boxIDList and not imageIDList:
            temp = spark.sparkContext.parallelize(boxIDList).map(lambda x : (x, )).toDF(["BoxID"])
            boxTemp = temp.join(bboxInfoTable, ["BoxID"], "inner")
            result = boxTemp.join(imageInfoTable, ["ImageID"], "inner")
        else:
            result = imageInfoTable
            
        #add BoxNumber column
        result = result.withColumn('BoxNumberPerImage', size('BoxIDList'))
        #add Area column
        result = result.withColumn('Area', result["Height"] * result["Width"])
    
        filtermask = ''
        imageIDListCate = []
        categoryList = None
        func = ImageSelector.__Numericfun
        if modifiedEdaFilterDict:
            edaNameList = [edaName for edaName in modifiedEdaFilterDict.keys()]
            for edaName, condit in modifiedEdaFilterDict.items():
                if edaName in numericSelection:
                    if len(condit) > 0:
                        filtermask += '('
                        for item in condit:
                            if type(item) == int:
                                filtermask += '{edaName} == {value} or '.format(edaName=edaName, value=item)
                            else:
                                filtermask += '({edaName} >= {left} and {edaName} <= {right}) or '.format(edaName=edaName, left=item[0], right=item[1])
                        filtermask = filtermask.rstrip('or ')
                        filtermask += ') and '
                elif edaName in categorySelection:
                    #Mode
                    if edaName == categorySelection[0]:
                        if len(condit) == 1:
                            filtermask += "{edaName} == '{conditList}' and " .format(edaName=edaName, conditList=condit[0])
                        else:
                            filtermask += '{edaName} in {conditList} and '.format(edaName=edaName, conditList=tuple(condit))
                    #Category
                    else:
                        # filter operateion. Don't generate filtermask
                        # filter categoryTable with categoryList and flat imageIDList of it
                        # and then join result 
                        categoryList = condit.copy()
                        imageIDListDf = categoryTable[categoryTable.Category.isin(condit)]
                        imageIDListDf = imageIDListDf.withColumn("ImageID", explode(imageIDListDf.ImageIDList))
                        result = imageIDListDf.join(result, ["ImageID"], "inner")
                else:
                    raise ValueError('The edaName %s is invalid.' % edaName)

            filtermask = filtermask.rstrip('and ')
            result.createOrReplaceTempView("result")
            result = spark.sql("select * from result where " + filtermask)
            if set(edaNameList)<= {'ImageLuminanceMean', 'ImageLuminanceStd'}:
                result = result[['ImageID','ImageLuminanceMean', 'ImageLuminanceStd']]
            elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
                result = result[['ImageID','Width', 'Height']]
            else:
                result = result[edaNameList+['ImageID']]
                for edaName, condit in modifiedEdaFilterDict.items():
                    if edaName in numericSelection:
                        numUdf = udf(lambda x : func(x, condit))
                        result = result.withColumn(edaName, numUdf(result[edaName]))
        return result, edaNameList, categoryList

    @multimethod(PdDataFrame, list, dict)
    def SearchBoxesOnEDA(self, bboxInfoTable, boxIDList, edaFilterList):
        """
        The function to search boxes by the edaFilterList.

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param boxIDList: a list of boxIDs(if None, means all boxes).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: search result, eda name list.
        :rtype: PdDataFrame, list
        """
        numericSelection = ['Area', 'LuminanceMean', 'LuminanceStd', 'Width', 'Height']
        categorySelection = ['Category']
        boolSelection = ['IsOccluded', 'IsTruncated', 'IsGroupOf', 'IsDepiction', 'IsInside']
    
        #check the edaFilterList
        modifiedEdaFilterList = self.__validateEDA(edaFilterList, numericSelection, categorySelection, boolSelection)
    
        result = None
        edaNameList = None
        
        if boxIDList:
            result = bboxInfoTable[bboxInfoTable.index.isin(boxIDList)]
        else:
            #This means user does not give any box ids then search on the bboxInfoTable.
            result = bboxInfoTable
    
        #add Width, Height column
        result = result.eval('Width = XMax - XMin', inplace=False)
        result.eval('Height = YMax - YMin', inplace=True)
    
        filtermask = ''
        if modifiedEdaFilterList:
            edaNameList = [edaName for edaName in modifiedEdaFilterList.keys()]
        
            for edaName, condit in modifiedEdaFilterList.items():
                if edaName in numericSelection:
                    if len(condit) > 0:
                        filtermask += '('
                        for item in condit:
                            if type(item) == int:
                                filtermask += '{edaName} == {value} or '.format(edaName=edaName, value=item)
                            else:
                                filtermask += '({edaName} >= {left} and {edaName} <= {right}) or '.format(edaName=edaName, left=item[0], right=item[1])
                        filtermask = filtermask.rstrip('or ')
                        filtermask += ') and '
                elif edaName in (categorySelection + boolSelection):
                    filtermask += '{edaName} in {conditList} and '.format(edaName=edaName, conditList=condit)
                else:
                    raise ValueError('The edaName %s is invalid.' % edaName)
                
            filtermask = filtermask.rstrip('and ')
            result = result.query(filtermask)

            if set(edaNameList)<= {'LuminanceMean', 'LuminanceStd'}:
                result = result[['LuminanceMean', 'LuminanceStd']]
            elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
                result = result[['Width', 'Height']]
            else:
                result = result[edaNameList]
                for edaName, condit in modifiedEdaFilterList.items():
                    if edaName in numericSelection:
                        result.loc[:, edaName] = result[edaName].map(lambda x: self.__Numericfun(x, condit))
                    
        return result, edaNameList
    
    @multimethod(SparkDataFrame, list, dict)
    def SearchBoxesOnEDA(self, bboxInfoTable, boxIDList, edaFilterDict):
        """
        The function to search boxes by the edaFilterList.

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param boxIDList: a list of boxIDs(if None, means all boxes).
        :type boxIDList: list
        :param edaFilterDict: a dict of eda filters[(edaname, filtercondition)].
        :type edaFilterDict: dict
        :return: search result, eda name list.
        :rtype: SparkDataFrame, list
        """
        numericSelection = ['Area', 'LuminanceMean', 'LuminanceStd', 'Width', 'Height']
        categorySelection = ['Category']
        boolSelection = ['IsOccluded', 'IsTruncated', 'IsGroupOf', 'IsDepiction', 'IsInside']
        spark = SparkSession.builder.getOrCreate()
        #check the edaFilterList
        modifiedEdaFilterDict = self.__validateEDA(edaFilterDict, numericSelection, categorySelection, boolSelection)
    
        result = None
        edaNameList = None
        func = ImageSelector.__Numericfun
        if boxIDList:
            temp = spark.sparkContext.parallelize(boxIDList).map(lambda x : (x, )).toDF(["BoxID"])
            result = temp.join(bboxInfoTable, ["BoxID"], "inner")
        else:
            #This means user does not give any box ids then search on the bboxInfoTable.
            result = bboxInfoTable
      
        #add Width, Height column
        result = result.withColumn('Width', result["XMax"] - result["XMin"])
        result = result.withColumn('Height', result["YMax"] - result["YMin"])

        filtermask = ''
        if modifiedEdaFilterDict:
            edaNameList = [edaName for edaName in modifiedEdaFilterDict.keys()]

            for edaName, condit in modifiedEdaFilterDict.items():
                if edaName in numericSelection:
                    if len(condit) > 0:
                        filtermask += '('
                        for item in condit:
                            if type(item) == int:
                                filtermask += '{edaName} == {value} or '.format(edaName=edaName, value=item)
                            else:
                                filtermask += '({edaName} >= {left} and {edaName} <= {right}) or '.format(edaName=edaName, left=item[0], right=item[1])
                        filtermask = filtermask.rstrip('or ')
                        filtermask += ') and '
                elif edaName in (categorySelection + boolSelection):
                    if len(condit) == 1:
                        filtermask += "{edaName} == '{conditList}' and ".format(edaName=edaName, conditList=condit[0])
                    else:
                        filtermask += '{edaName} in {conditList} and '.format(edaName=edaName, conditList=tuple(condit))
                else:
                    raise ValueError('The edaName %s is invalid.' % edaName)
        
            filtermask = filtermask.rstrip('and ')
            result.createOrReplaceTempView("result")
            result = spark.sql("select * from result where " + filtermask)
        
            if set(edaNameList)<= {'LuminanceMean', 'LuminanceStd'}:
                result = result[['BoxID','LuminanceMean', 'LuminanceStd']]
            elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
                result = result[['BoxID','Width', 'Height']]
            else:
                result = result[edaNameList + ["BoxID"]]
                for edaName, condit in modifiedEdaFilterDict.items():
                    if edaName in numericSelection:
                        numUdf = udf(lambda x : func(x, condit))
                        result = result.withColumn(edaName, numUdf(result[edaName]))
        return result, edaNameList

    @multimethod(PdDataFrame, PdDataFrame, PdDataFrame, list, dict)
    def AnalyzeImages(self, imageInfoTable, bboxInfoTable, categoryTable, imageIDList, edaFilterList):
        """
        Analyze images depending on EDA metrics and will return different kind of metrics:
        (1) Grayscale or color image distribution
        (2) Bounding boxes number distribution
        (3) Image size distribution
        (4) Image category distribution
        (5) Image luminance distribution
        (6) Image quality distribution (images with low night noise distribution)
        (7) ...

        :param imageInfoTable: image info table.
        :type imageInfoTable: PdDataFrame
        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param categoryTable: category table.
        :type categoryTable: PdDataFrame
        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: the analyze result.
        :rtype: ImageAnalyzeResult
        """

        rdic = {}
        target = None
        result, edaNameList, _ = self.SearchImagesOnEDA(imageInfoTable, bboxInfoTable, categoryTable, imageIDList, [], edaFilterList)
    
        if result is None or result.shape[0] == 0 or not edaNameList:
            return ImageAnalyzeResult(self.env, rdic)
    
        result.reset_index(inplace=True)
        if set(edaNameList)<= {'ImageLuminanceMean', 'ImageLuminanceStd'}:
            #Luminance distribution 
            target = 'Luminance'
            rdic['Luminance'] = dict(zip(result['ImageLuminanceMean'].tolist(), result['ImageLuminanceStd'].tolist()))
        elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
            #Area distribution 
            target = 'Area'
            rdic['Area'] = dict(zip(result['Width'].tolist(), result['Height'].tolist()))
        else:   
            #combination distribution
            target = 'Distribution'
            series = result.groupby(edaNameList, observed=True)['ImageID'].count()
            series.sort_values(ascending=True, inplace=True)
            keyList = series.index.tolist()
            for i, row in enumerate(keyList):
                key = ''
                if type(row) == str:
                    key = row
                else:
                    key = ','.join(row)
                keyList[i] = key
            rdic['Distribution'] = OrderedDict(zip(keyList, series.tolist()))
    
        return ImageAnalyzeResult(self.env, rdic, target, edaNameList) 
    
    @multimethod(SparkDataFrame, SparkDataFrame, SparkDataFrame, list, dict)
    def AnalyzeImages(self, imageInfoTable, bboxInfoTable, categoryTable, imageIDList, edaFilterDict):
        """
        Analyze images depending on EDA metrics and will return different kind of metrics:
        (1) Grayscale or color image distribution
        (2) Bounding boxes number distribution
        (3) Image size distribution
        (4) Image category distribution
        (5) Image luminance distribution
        (6) Image quality distribution (images with low night noise distribution)
        (7) ...

        :param imageInfoTable: image info table.
        :type imageInfoTable: SparkDataFrame
        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param categoryTable: category table.
        :type categoryTable: SparkDataFrame
        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param edaFilterDict: a dict of eda filters[(edaname, filtercondition)].
        :type edaFilterDict: dict
        :return: the analyze result.
        :rtype: ImageAnalyzeResult
        """
        rdic = {}
        target = None
        # result : imageInfoTable
        result, edaNameList, _ = self.SearchImagesOnEDA(imageInfoTable, bboxInfoTable, categoryTable, imageIDList, [], edaFilterDict)
        
        if result is None or result.count() == 0 or len(result.columns) == 0 or not edaNameList:
            return ImageAnalyzeResult(self.env, rdic)
        
        if set(edaNameList)<= {'ImageLuminanceMean', 'ImageLuminanceStd'}:
            #Luminance distribution
            resultCollect = result.select('ImageLuminanceMean', 'ImageLuminanceStd').collect()
            target = 'Luminance'
            rdic['Luminance'] = dict(zip([row["ImageLuminanceMean"] for row in resultCollect], [row["ImageLuminanceStd"] for row in resultCollect]))
        elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
            #Area distribution 
            resultCollect = result.select('Width', 'Height').collect()
            target = 'Area'
            rdic['Area'] = dict(zip([row["Width"] for row in resultCollect], [row["Height"] for row in resultCollect]))
        else:
            #combination distribution
            target = 'Distribution'
            df = result.groupby(edaNameList).count().sort("count")
            resultCollect = df.collect()
            keyList = []
            for row in resultCollect:
                element = ""
                for key in edaNameList:
                    element += str(row[key]) + ", "
                element = element.rstrip(", ")
                keyList.append(element)
            rdic['Distribution'] = OrderedDict(zip(keyList, [row["count"] for row in resultCollect]))
        return ImageAnalyzeResult(self.env, rdic, target, edaNameList) 

    @multimethod(PdDataFrame, list, dict)
    def AnalyzeImagesByBoundingBoxes(self, bboxInfoTable, boxIDList, edaFilterList):
        """
        Analyze bounding boxes depending on EDA metrics and will return different kind of metrics:
        (1) bb-level category distribution
        (2) bb-level attributes distribution like isOccluded bbs distribution, isTruncated bbs distribution, isGrouped bbs distribution, ...
        (3) bb size distribution
        (4) bb luminance distribution
        (5) ...

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: the analyze result.
        :rtype: ImageAnalyzeResult
        """
        rdic = {}
        target = None
        result, edaNameList = self.SearchBoxesOnEDA(bboxInfoTable, boxIDList, edaFilterList)
      
        if result is None or result.shape[0] == 0 or not edaNameList:
            return ImageAnalyzeResult(self.env, rdic) 
    
        result.reset_index(inplace=True)
        if set(edaNameList)<= {'LuminanceMean', 'LuminanceStd'}:
            #Luminance distribution 
            target = 'Luminance'
            rdic['Luminance'] = dict(zip(result['LuminanceMean'].tolist(), result['LuminanceStd'].tolist()))
        elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
            #Area distribution 
            target = 'Area'
            rdic['Area'] = dict(zip(result['Width'].tolist(), result['Height'].tolist()))
        else:   
            #combination distribution
            target = 'Distribution'
            series = result.groupby(edaNameList, observed=True)['BoxID'].count()
            series.sort_values(ascending=True, inplace=True)
            keyList = series.index.tolist()
            for i, row in enumerate(keyList):
                key = ''
                if type(row) == str:
                    key = row
                elif type(row) == int:
                    key = str(row)
                else:
                    key = ','.join([str(el) for el in row])
                keyList[i] = key
            rdic['Distribution'] = OrderedDict(zip(keyList, series.tolist()))
    
        return ImageAnalyzeResult(self.env, rdic, target, edaNameList)   

    @multimethod(SparkDataFrame, list, dict)
    def AnalyzeImagesByBoundingBoxes(self, bboxInfoTable, boxIDList, edaFilterDict):
        """
        Analyze bounding boxes depending on EDA metrics and will return different kind of metrics:
        (1) bb-level category distribution
        (2) bb-level attributes distribution like isOccluded bbs distribution, isTruncated bbs distribution, isGrouped bbs distribution, ...
        (3) bb size distribution
        (4) bb luminance distribution
        (5) ...

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterDict: a dict of eda filters[(edaname, filtercondition)].
        :type edaFilterDict: dict
        :return: the analyze result.
        :rtype: ImageAnalyzeResult
        """
        rdic = {}
        target = None
        result, edaNameList = self.SearchBoxesOnEDA(bboxInfoTable, boxIDList, edaFilterDict)
    
        if result is None or result.count() == 0 or not edaNameList:
            return ImageAnalyzeResult(self.env, rdic) 
    
        if set(edaNameList)<= {'LuminanceMean', 'LuminanceStd'}:
            #Luminance distribution 
            resultCollect = result.select('LuminanceMean', 'LuminanceStd').collect()
            target = 'Luminance'
            rdic['Luminance'] = dict(zip([row["LuminanceMean"] for row in resultCollect], [row["LuminanceStd"] for row in resultCollect]))
        elif set(edaNameList) <= {'Width', 'Height', 'Area'}:
            #Area distribution 
            target = 'Area'
            resultCollect = result.select('Width', 'Height').collect()
            rdic['Area'] = dict(zip([row["Width"] for row in resultCollect], [row["Height"] for row in resultCollect]))
        else:
            #combination distribution
            target = 'Distribution'
            df = result.groupby(edaNameList).count().sort("count")
            resultCollect = df.collect()
            keyList = []
            for row in resultCollect:
                element = ""
                for key in edaNameList:
                    element += str(row[key]) + ", "
                element = element.rstrip(", ")
                keyList.append(element)
            rdic['Distribution'] = OrderedDict(zip(keyList, [row["count"] for row in resultCollect]))
        return ImageAnalyzeResult(self.env, rdic, target, edaNameList)
    
    @multimethod(PdDataFrame, PdDataFrame, PdDataFrame, list, list, dict)
    def SelectImages(self, imageInfoTable, bboxInfoTable, categoryTable, imageIDList, boxIDList, edaFilterList):
        """
        Select images
        People can select images on different rules and return image id list:
        1. select images by ids, if none, select all images.
        2. select images by one or multiple combination of the following image level eda metrics: 
            (1) image-level category, 
            (2) image size, 
            (3) bounding box number, 
            (4) bounding box ids, 
            (5) image luminance
            (6) image quality (low night noise)
            (7) Grayscale or color image
            (8) ...

        :param imageInfoTable: image info table.
        :type imageInfoTable: PdDataFrame
        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param categoryTable: category table.
        :type categoryTable: PdDataFrame
        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: a dict like {ImageID: [bboxID1, bboxID2,...]}.
        :rtype: dict
        """
        result, _, categoryList = self.SearchImagesOnEDA(imageInfoTable, bboxInfoTable, categoryTable, imageIDList, boxIDList, edaFilterList)
    
        if not categoryList and not boxIDList:
            result = imageInfoTable[imageInfoTable.index.isin(result.index.tolist())]
            return dict(zip(result.index.tolist(), result['BoxIDList'].tolist()))
        else:
            result = bboxInfoTable[bboxInfoTable['ImageID'].isin(result.index.tolist())]
            if categoryList:
                result = result[result['Category'].isin(categoryList)]
            if boxIDList:      
                 result = result[result.index.isin(boxIDList)]
        
            result = result.reset_index().groupby('ImageID', observed=True)['BoxID'].apply(list)
            return dict(zip(result.index.tolist(), result.tolist()))

    @multimethod(SparkDataFrame, SparkDataFrame, SparkDataFrame, list, list, dict)
    def SelectImages(self, imageInfoTable, bboxInfoTable, categoryTable, imageIDList, boxIDList, edaFilterDict):
        """
        Select images
        People can select images on different rules and return image id list:
        1. select images by ids, if none, select all images.
        2. select images by one or multiple combination of the following image level eda metrics: 
            (1) image-level category, 
            (2) image size, 
            (3) bounding box number, 
            (4) bounding box ids, 
            (5) image luminance
            (6) image quality (low night noise)
            (7) Grayscale or color image
            (8) ...

        :param imageInfoTable: image info table.
        :type imageInfoTable: SparkDataFrame
        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param categoryTable: category table.
        :type categoryTable: SparkDataFrame
        :param imageIDList: a list of imageIDs(if None, means all images).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterDict: a dict of eda filters[(edaname, filtercondition)].
        :type edaFilterDict: dict
        :return: a dict like {ImageID: [bboxID1, bboxID2,...]}.
        :rtype: dict
        """
        result, _, categoryList = self.SearchImagesOnEDA(imageInfoTable, bboxInfoTable, categoryTable, imageIDList, boxIDList, edaFilterDict)
        spark = SparkSession.builder.getOrCreate()
        if not categoryList and not boxIDList:
            temp = result.select("ImageID")
            result = temp.join(imageInfoTable,["ImageID"], "inner")
            return dict(zip(result.select("ImageID").rdd.map(lambda x : x[0]).collect(), result.select("BoxIDList").rdd.map(lambda x : x[0]).collect()))
        else:
            temp = result.select("ImageID")
            result = temp.join(bboxInfoTable,["ImageID"], "inner")
            if categoryList:
                result = result[result['Category'].isin(categoryList)]
            if boxIDList:
                temp = spark.sparkContext.parallelize(boxIDList).map(lambda x : (x, )).toDF(["BoxID"])
                result = temp.join(result,["BoxID"], "inner")
            result = result.dropDuplicates().groupby('ImageID').agg(F.collect_list(F.col("BoxID")).alias("BoxList"))
            boxCollect = result.select("BoxList").rdd.map(lambda x : x[0]).collect()
            imageCollect = result.select("ImageID").rdd.map(lambda x : x[0]).collect()
            return dict(zip(imageCollect, boxCollect))

    @multimethod(PdDataFrame, list, dict)
    def SelectBboxs(self, bboxInfoTable, boxIDList, edaFilterList):
        """
        Select bounding boxes (bbs)
        People can select bbs on different rules and bbs id list:
        1. select bbs by ids, if none, select all bbs
        2. select bbs by one or multiple combination of the following bb level eda metrics:
            (1) bb-level category
            (2) bb-level attributes like isOccluded, isTruncated, isGrouped, ...
            (3) bb size
            (4) bb luminace
            (5) ...

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterList: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterList: dict
        :return: a boxID list of the query results.
        :rtype: list
        """
        result, _ = self.SearchBoxesOnEDA(bboxInfoTable, boxIDList, edaFilterList)
        return result.index.tolist()

    @multimethod(SparkDataFrame, list, dict)
    def SelectBboxs(self, bboxInfoTable, boxIDList, edaFilterDict):
        """
        Select bounding boxes (bbs)
        People can select bbs on different rules and bbs id list:
        1. select bbs by ids, if none, select all bbs
        2. select bbs by one or multiple combination of the following bb level eda metrics:
            (1) bb-level category
            (2) bb-level attributes like isOccluded, isTruncated, isGrouped, ...
            (3) bb size
            (4) bb luminace
            (5) ...

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param edaFilterDict: a dict of eda filters{edaname: filtercondition}.
        :type edaFilterDict: dict
        :return: a boxID list of the query results.
        :rtype: list
        """
        result, _ = self.SearchBoxesOnEDA(bboxInfoTable, boxIDList, edaFilterDict)
        return result.select("BoxID").rdd.map(lambda x : x[0]).collect()
        
    @multimethod(PdDataFrame)
    def GetCategories(self, categoryTable):
        """
        Get all categories.

        :param categoryTable: category table.
        :type categoryTable: PdDataFrame
        :return: a category list.
        :rtype: list
        """
        return categoryTable.index.tolist()

    @multimethod(SparkDataFrame)
    def GetCategories(self, categoryTable):
        """
        Get all categories.

        :param categoryTable: category table.
        :type categoryTable: SparkDataFrame
        :return: a category list.
        :rtype: list
        """
        return categoryTable.select("Category").rdd.map(lambda x : x[0]).collect()

    @multimethod(PdDataFrame, PdDataFrame, list, list, str, str, int)
    def ShowImages(self, bboxInfoTable, urlTable, imageIDList, boxIDList, urlPrefix, path, limitedNum):
        """
        Show bounding boxes in images and will download images with drawn bounding boxes to local path. 

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: PdDataFrame
        :param urlTable: image url table.
        :type urlTable: PdDataFrame
        :param imageIDList: a list of imageIDs(if None, select image id by boxIDList).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all bbox).
        :type boxIDList: list
        :param urlPrefix: the download url prefix.
        :type urlPrefix: str
        :param path: local save path.
        :type path: str
        :param limitedNum: the download image number.
        :type limitedNum: int
        """

        if not os.path.exists(path):
            os.makedirs(path)
        
        if not boxIDList and not imageIDList:
            raise ValueError("Input can not be empty.")  
            
        if boxIDList:
            bboxInfoTable = bboxInfoTable[bboxInfoTable.index.isin(boxIDList)]
        if not imageIDList:
            df = bboxInfoTable.drop_duplicates(subset = ["ImageID"])
            imageIDList = df["ImageID"].tolist()
    
        count = 0
        print('download start...')
        for imageId in imageIDList: 
            if count == limitedNum:
                break;
            url = urlPrefix + urlTable.loc[imageId, 'Subset'] + "/" + imageId + ".jpg"
            req = urlopen(url)
            arr = np.asarray(bytearray(req.read()), dtype=np.uint8)
            img = cv2.imdecode(arr, -1)
            multiBox = bboxInfoTable[bboxInfoTable["ImageID"] == imageId]
            #if no box, continue
            if multiBox.shape[0]==0:
                continue
            labelLists = multiBox["Category"].tolist()
            xminLists = multiBox["XMin"].tolist()
            xmaxLists = multiBox["XMax"].tolist()
            yminLists = multiBox["YMin"].tolist()
            ymaxLists = multiBox["YMax"].tolist()
            for i in range(len(labelLists)):
                cv2.rectangle(img, (xminLists[i], yminLists[i]), (xmaxLists[i],ymaxLists[i]), (0,255,0), 2)
                text = labelLists[i]
                cv2.putText(img, text, (xminLists[i], max(yminLists[i],10)), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0,0,255), 1)
            cv2.imwrite(os.path.join(path, imageId + ".jpg"), img)
            count += 1

        print('finish.')

    @multimethod(SparkDataFrame, dict, list, list, str, str, int)
    def ShowImages(self, bboxInfoTable, urlDict, imageIDList, boxIDList, urlPrefix, path, limitedNum):
        """
        Show bounding boxes in images and will download images with drawn bounding boxes to dbfs path. 

        :param bboxInfoTable: bounding box info table.
        :type bboxInfoTable: SparkDataFrame
        :param urlDict: image url table.
        :type urlDict: dict
        :param imageIDList: a list of imageIDs(if None, select image id by boxIDList).
        :type imageIDList: list
        :param boxIDList: a list of boxIDs(if None, means all boxes).
        :type boxIDList: list
        :param urlPrefix: the download url prefix.
        :type urlPrefix: str
        :param path: dbfs save path.
        :type path: str
        :param limitedNum: the download image number.
        :type limitedNum: int
        """

        if not boxIDList and not imageIDList:
            raise ValueError("Input can not be empty.")
        if path.startswith("dbfs:"):
            path = path.split(":")[1]
        elif path.startswith("/"):
            path = path
        else:
            raise ValueError("The path is invalid")
        
        spark = SparkSession.builder.getOrCreate()
        if boxIDList:
            temp = spark.sparkContext.parallelize(boxIDList).map(lambda x : (x, )).toDF(["BoxID"])
            bboxInfoTable = temp.join(bboxInfoTable,["BoxID"], "inner")
        if imageIDList:
            temp = spark.sparkContext.parallelize(imageIDList).map(lambda x : (x, )).toDF(["ImageID"])
            bboxInfoTable = temp.join(bboxInfoTable,["ImageID"], "inner")
        func = ImageSelector.showBoxInImage
        result = bboxInfoTable.withColumn("BoxList", F.struct(F.col("XMin"),F.col("XMax"),F.col("YMin"),F.col("YMax"),F.col("Category")))
        result = result.groupBy("ImageID").agg(F.collect_list("BoxList").alias("BoxList")).select("ImageID", "BoxList")
        urlUdf = udf(lambda x : urlDict[x] + "/" + x)
        # result : "ImageID": subset/imageId.jpg, "BoxList" : ("XMin")
        result = result.withColumn("ImageID", urlUdf(result["ImageID"]))

        if not limitedNum:
            result = result.limit(limitedNum)
        result.rdd.foreachPartition(lambda x : func(urlPrefix,path,x))
    